define [
	'underscore'
	'config'
	'model'
], (_, Config, Model) ->

	class BlockModel extends Model

		url: ->
			_.template Config.urls.block, id: @get('id')

		defaults:
			name: ''
			price: ''
			scope: null
			time: 0
			duration: 30
			invalid: no
			editable: yes
			movable: yes
			stretchable: no
			stretched_by: []
			canceled: no
			finished: no
			color: '#e37f00'

		getters:
			left: ->
				@get('time')*Config.baseWidth

			width: ->
				duration = @get('fullDuration')*Config.baseWidth

			fullDuration: ->
				duration = @get('duration')
				stretched_by = @get('stretched_by') or []
				for s in stretched_by
					block = @collection.get(s)
					duration+= block.get('duration') if block
				duration

		setters:
			left: (value) ->
				value = Math.max value, 0
				@set 'time', Math.round value/Config.baseWidth
				undefined

			width: (value) ->
				@set 'duration', Math.round value/Config.baseWidth
				undefined

			duration: (value) ->
				return Math.max 1, parseInt(value, 10)

			time: (value) ->
				return Math.max 0, parseInt(value, 10)


		initialize: ->
			@on 'checkStretched', ->
				# log "Block on checkStretched", @id
				scope = @get('scope')
				old = @toJSON()
				stretched_by = []

				@set
					stretched_by: []
				, silent: yes

				@collection.sort().search({scope}).each (block, index) =>
					my =
						id: @id
						cid: @cid
						duration: @get('fullDuration')
						time: @get('time')

					enemy =
						id: block.id
						cid: block.cid
						time: block.get('time'),
						duration: block.get('fullDuration')

					# To jest ten sam blok co ja lub jest też stretchable, a tak być nie może,
					# a może jeszcze nie ma ID! no normalnie masakra
					return if my.cid is enemy.cid or block.get('stretchable') or not block.id?

					if @_checkCollision(my, enemy)
						stretched_by.push enemy.id
					else
						stretched_by = _.without stretched_by, enemy.id

					@set
						stretched_by: stretched_by
					, silent: yes

				# Wywalamy ewentualne duplikaty i sortujemy
				stretched_by = _.uniq stretched_by
				stretched_by.sort()

				# Jak się nie zmieniła to idziemy stąd
				return if _.isEqual stretched_by, old.stretched_by

				# zapisujemy zmiany
				@trigger("change")
				@save
					stretched_by: stretched_by
				, checkStretchable: off, info: 'after checkStretched'

			@on 'change:id', =>
				@isValid()

			@on 'endEdit', =>
				@save() unless @isNew()



		validate: (attrs = {}, options = {}) ->
			# log "BlockModel::validate", @id

			error = null
			isFine = yes

			my =
				id: @id
				cid: @cid
				scope: @get('scope')
				time: @get('time')
				duration: @get('fullDuration')
				stretchable: @get('stretchable')

			error = "no scope" unless my.scope

			if my.time+my.duration > Config.max
				error = "out of range"
				isFine = no

			# Nie jest w żadnej kolekcji, idziemy stąd
			return error unless @collection

			###
				Sprawdzamy kolizję
			###
			isCollision = no
			@collection.sort().each (block, index) =>

				# Uciekamy jak te same ID
				return if my.cid is block.cid

				enemy =
					scope: block.get('scope')
					time: block.get('time')
					duration: block.get('fullDuration')
					stretchable: block.get('stretchable')

				# ten sam scope czyli możliwa kolizja
				if enemy.scope is my.scope and @_checkCollision(my, enemy)
					# Jeśli obydwoje mają strecza lub nie mają to jest kolizja
					if (enemy.stretchable and my.stretchable) or (not enemy.stretchable and not my.stretchable)
						isCollision = yes
						# return

			if isCollision
				isFine = no
				error = "collision"


			###
				Jak nie ma kolizji to sprawdzamy czy jest streczable
				chyba, że mamy tego nie sprawdzać, bo zapętlanie jest średnio fajne
			###
			if not isCollision and options.checkStretchable isnt off

				# Ja jestem rozciągliwy, trzeba sprawdzić czy wszystko ok
				if my.stretchable
					@trigger 'checkStretched'

				# Lecimy po innych i szukamy problemu
				@collection.each (block, index) =>

					# Uciekamy jak te same ID
					return if my.cid is block.cid

					enemy =
						scope: block.get('scope')
						time: block.get('time')
						duration: block.get('fullDuration')
						stretchable: block.get('stretchable')

					# ten sam scope czyli możliwa kolizja
					if enemy.scope is my.scope and @_checkCollision(my, enemy)
						# enemy jest rozciągliwy
						if enemy.stretchable
							block.trigger 'checkStretched'

					# Pamiętamy, że możemy wyciągnać klocka z środka,
					# więc dla pozostałych też trzeba sprawdić
					else
						if enemy.stretchable
							block.trigger 'checkStretched'

					# Teraz sprawdzamy czy wystąpił jakiś problem ze checkStretched
					if block.get('invalid') and my.id in block.get('stretched_by')
						error = "siblings problem"
						isFine = no
						# return # no i kończymy zabawę w szukanie

			@set invalid: not isFine
			error

		save: (attributes, options) ->
			# log "BlockModel::save", @id
			result = super

			# To już po próbie zapisu, zwracamy, żeby się nie zapętlić
			return result if options?.invoked

			# Próbujemy zapisać pozostałe błędne
			@collection?.each (block) =>
				if block.id isnt @id and block.get('invalid')
					block.save null, invoked: yes

			result

		_checkCollision: (a, b) ->
			(a.time >= b.time and a.time < b.time+b.duration) or
			(a.time+a.duration <= b.time+b.duration and a.time+a.duration > b.time) or
			(a.time <= b.time and a.time+a.duration > b.time+b.duration) # wewnątrz mamy coś
